// ************************************************************************************************
//
// Nixie-Uhr unter Verwendung des
//
// PIC 16F627A
//
// Ausgabe der Zeit ber 6 russische IN-17 Nixies, Einstellen der Zeit ber einen Taster
//
// Dimmen der Nixies ber seperaten Taster
//
// Copyright www.jb-electronics.de
//
// ************************************************************************************************

// PIC-Header einbinden
#include <pic.h>

// ************************************************************************************************
// Methodenprototypen
void updateDisplay ();
void schreibeWert (uchar Wert, uchar Wohin);
uchar divide (uchar was, uchar womit, uchar modus);
void updatePWM ();
void advanceTime ();
void AntiCathodePoisoning (uchar Count);
void warte ();

// ************************************************************************************************
// Globale Variablen
static uchar PWMLevel, ps, Sekunden, Minuten, Stunden;
static uchar SekundenWert, MinutenWert, StundenWert;
static uchar SW1_BUFFER, SW2_BUFFER, SW3_BUFFER, SW4_BUFFER;
static uchar ACP[10] = {5, 4, 9, 3, 6, 0, 8, 2, 7, 1};
static int PWMTable[8] = {0, 140, 181, 256, 362, 512, 724, 1023};
static int DUTY_CYCLE, TimeCounter;

// 12- oder 24-Stunden Format
static unsigned char max_hours = 24;

// ************************************************************************************************
// Hauptmethode
void main (void) {

    // ********************************************************************************************
	// Konfigurationsbits setzen
	__CONFIG(0x3F2A);


	// ********************************************************************************************
	// Portspezifikationen

	// PORTB als Ausgang definieren
	TRISB = 0b00000000;

	// PORTB unterteilen
	#define STROBE_SEKUNDEN RB6
	#define STROBE_MINUTEN	RB1
	#define STROBE_STUNDEN	RB2
	#define DATA			RB5
	#define CLOCK			RB4
	#define NIXIE_ENABLE	RB3


	// PORTA als Eingang definieren
	TRISA = 0b1111;
	CMCON = 0b111;
	
	// PORTA unterteilen
	#define ANTICATHODE		RA0
	#define STELLEN_MIN		RA1
	#define STELLEN_STD		RA2
	#define NIXIES_DIMM		RA3


	// ********************************************************************************************
	// TIMER setzen (Interrupt-Routine wird so genau 4096mal pro Sekunde aufgerufen)
	T0CS = 0;
	PSA = 1;
	PS0 = 1;
	PS1 = 1;
	PS2 = 1;

	// Interrupt bei Timer-berlauf auslsen
	T0IE = 1;

	// Interrupts aktivieren
	GIE = 1;


	// ********************************************************************************************
	// PWM-Settings:

	// Obergrenze von Timer2 setzen:
	PR2 = 0xff; // 255

	// Duty cycle festlegen (0...1023)
	// Setzen der Register in Methode updatePWM()
	DUTY_CYCLE = 0;

	// CCP1 als Ausgang
	TRISB3 = 0;

	// Prescaler von Timer2 auf 1:1 setzen (00 = 1:1, 01 = 1:4, 1x = 1:16)
	// und Timer einschalten (Bit Nr. 2)
	T2CON = 0b100;

	// Das CCP1 Modul auf PWM setzen
	CCP1CON = 0b1100;

	// PWM-Level setzen (0...7)
	PWMLevel = 7;
	ps = -1;

	// Startwerte
	Stunden = 12;
	Minuten = 34;
	Sekunden = 56;
	

	// ********************************************************************************************
	// Beginn der Hauptschleife
	while (true) {

		updateDisplay ();

		DUTY_CYCLE = PWMTable[PWMLevel];
		updatePWM ();

		// Uhrzeit anzeigen
		StundenWert = Stunden;
		MinutenWert = Minuten;
		SekundenWert = Sekunden;

		// Minuten einstellen?
		if ((SW1_BUFFER == 0) && (STELLEN_MIN)) {
			SW1_BUFFER = 100;
			Sekunden = 0;
			TimeCounter = 0;
			Minuten++;
			if (Minuten >= 60) {
				Minuten = 0;
			}
		} else if (STELLEN_MIN) {
			SW1_BUFFER = 100;
		}

		// Stunden einstellen?
		if ((SW2_BUFFER == 0) && (STELLEN_STD)) {
			SW2_BUFFER = 100;
			Sekunden = 0;
			TimeCounter = 0;
			Stunden++;
			if (Stunden >= max_hours) {
				Stunden = 0;
			}
		} else if (STELLEN_STD) {
			SW2_BUFFER = 100;
		}

		// Helligkeit einstellen?
		if ((SW3_BUFFER == 0) && (NIXIES_DIMM)) {
			SW3_BUFFER = 100;
			PWMLevel = PWMLevel + ps;
			if ((PWMLevel >= 7) || (PWMLevel <= 0)) {
				ps = -ps;
			}
		} else if (NIXIES_DIMM) {
			SW3_BUFFER = 100;
		}

		// Anti-Cathode-Poisoning-Routine?
		if ((SW4_BUFFER == 0) && (ANTICATHODE)) {
			SW4_BUFFER = 500;
			AntiCathodePoisoning (20);
		} else if (ANTICATHODE) {
			SW1_BUFFER = 500;
		}

	}

}

// ************************************************************************************************
// Interrupt-Routine fr genauen 1Hz-Takt
static void interrupt isr  (void) {

	// Eine Sekunde um? (wird 4096 mal pro Sekunde ausgefhrt)
	if (T0IF) {

		TimeCounter++;
		if (TimeCounter >= 4095) {
			advanceTime ();
			TimeCounter = 0;
		}

		// Taster entprellen
		if (SW1_BUFFER > 0) {
			SW1_BUFFER--;
		}
		if (SW2_BUFFER > 0) {
			SW2_BUFFER--;
		}
		if (SW3_BUFFER > 0) {
			SW3_BUFFER--;
		}
		if (SW4_BUFFER > 0) {
			SW4_BUFFER--;
		}

		T0IF = 0;

	}

}

// ************************************************************************************************
// Routine zum Anzeigen des Displays
void updateDisplay () {

	schreibeWert (SekundenWert, 1);
	schreibeWert (MinutenWert,  2);
	schreibeWert (StundenWert,  3);

}

// ************************************************************************************************
// schreibt einen Wert in ein Schieberegister
void schreibeWert (uchar Wert, uchar Wohin) {

	// Variablen
	int Index;
	uchar Einer, Zehner;

	// initialisieren
	STROBE_SEKUNDEN = 0;
	STROBE_MINUTEN = 0;
	STROBE_STUNDEN = 0;

	// Einer und Zehner ermitteln
	Zehner = Wert / 10;
	Einer = Wert % 10;

	// erst die Zehner
	for (Index = 3; Index >= 0; Index--) {
		NOP();
		if (Zehner & (1 << Index)) {
			DATA = 1;
		} else {
		 	DATA = 0;
		}
		NOP();
		CLOCK = 1;
		NOP();
		CLOCK = 0;
	}

	// dann die Einer
	for (Index = 3; Index >= 0; Index--) {
		NOP();
		if (Einer & (1 << Index)) {
			DATA = 1;
		} else {
		 	DATA = 0;
		}
		NOP();
		CLOCK = 1;
		NOP();
		CLOCK = 0;
	}

	// Wert bernehmen
	if (Wohin == 1) {
		STROBE_SEKUNDEN = 1;
	} else if (Wohin == 2) {
		STROBE_MINUTEN = 1;
	} else if (Wohin == 3) {
		STROBE_STUNDEN = 1;
	}
	NOP();

}

// ************************************************************************************************
// Diese Methode setzt den neuen DUTY_CYCLE Wert fest 
void updatePWM () {

	CCPR1L  = DUTY_CYCLE >> 2;
	CCP1CON = 0b1100 | ((DUTY_CYCLE & 0b11) << 2);

}

// ************************************************************************************************
// Diese Methode zhlt die Zeitinformationen einen Schritt weiter
void advanceTime () {

	Sekunden++;
	if (Sekunden >= 60) {
		Sekunden = 0;
		Minuten++;
		if (Minuten >= 60) {
			Minuten = 0;
			Stunden++;
			if (Stunden >= max_hours) {
				Stunden = 0;
			}
		}
	}

}

// ************************************************************************************************
// Diese Methode unterdrckt das Cathode-Poisoning
void AntiCathodePoisoning (uchar Count) {

	uchar Counter1, Counter2, c1, c2;
	for (Counter1 = 0; Counter1 < Count; Counter1++) {
		for (Counter2 = 0; Counter2 < 10; Counter2++) {
			SekundenWert =  ACP[9 - Counter2] + ACP[Counter2] * 10;
			MinutenWert  =  ACP[9 - Counter2] + ACP[Counter2] * 10;
			StundenWert  =  ACP[9 - Counter2] + ACP[Counter2] * 10;
			updateDisplay ();
			for (c1 = 0; c1 < 255; c1++) {
				for (c2 = 0; c2 < 10; c2++ ) {
					warte();
				}
			}
		}
		for (Counter2 = 8; Counter2 > 1; Counter2--) {
			SekundenWert =  ACP[9 - Counter2] + ACP[Counter2] * 10;
			MinutenWert  =  ACP[9 - Counter2] + ACP[Counter2] * 10;
			StundenWert  =  ACP[9 - Counter2] + ACP[Counter2] * 10;
			updateDisplay ();
			for (c1 = 0; c1 < 255; c1++) {
				for (c2 = 0; c2 < 10; c2++ ) {
					warte();
				}
			}
		}
	}

}

// ************************************************************************************************
// Diese Methode wartet einen Takt
void warte () {

	NOP();

}
